/**********************************************************************
  CGHseg_rewrite.c
  Weil Lai

  July 2007

  The motivation behind rewriting the code is to reduce memory
  consumption by the generation of the cost matrix, use R's memory
  allocation routines, and convert the code to use zero-based indexing.
**********************************************************************/

#include <R.h>

/**********************************************************************
   Homoscedastic model of Picard et al.
   Selection of optimal # of segments is done in the R code

   Input:
      - N       integer of length 1 (specifies length of Y1)
      - Y1      double vector of length N (contains input data)
      - Km      integer of length 1 (specifies the max # segments to find)

   Output:
      - J       double vector of length Km (log-likelihood)
      - th      integer vector of length Km*Km (matrix of breakpoints)
**********************************************************************/
void CGHseg3_th(double *N1, double *Y1, int *Km, double *J, int *th)
{
  /* declare variables by order of appearance */
  long i, j, k, L, minIndex, K;
  long *nrV, *t;
  long N;

  double temp, minVal;
  double *Y2, *Y1i, *Y2i, *Y2i_new, *Y1i_new;
  double *G, *I;

  /* Since only the upper diagonal of G is used, we can save memory by
     initializing only N*(N+1)/2 elements instead of N*N.  This becomes
     very important for very large data sets.  nrV is an intermediate variable
     used for keeping track of the number of elements per column
     in the upper diagonal matrix */
  N=(long) *N1;
  G = (double *) Calloc((N)*(N+1)/2, double);
  nrV = (long *) Calloc(N, long);

  for(i = 1; i < N; i++)
    nrV[i] = nrV[i-1] + i;

  /* initialize I and t matrices */
  I = (double *) Calloc((*Km)*(N), double);
  t = (long *) Calloc((*Km-1)*(N), long);

  /* Initialize Y2, Y1i, Y2i */
  Y2 = (double *) Calloc(N, double);
  Y1i = (double *) Calloc(N, double);
  Y2i = (double *) Calloc(N, double);

  for(i = 0; i < N; i++)
    Y2[i] = Y1[i] * Y1[i];

  /* Calculate culmulative sums */
  Y1i[0] = Y1[0];
  Y2i[0] = Y2[0];
  for(i = 1; i < N; i++)  {
    Y1i[i] = Y1[i] + Y1i[i-1];
    Y2i[i] = Y2[i] + Y2i[i-1];
  }

  /* R code: G[1, ] <- Y2i - (Y1i*Y1i)/(1:n) */
  for(j = 0; j < N; j++)
    G[0 + nrV[j]] = Y2i[j] - Y1i[j]*Y1i[j]/(j+1);

  /* R code: for( i in 2:n )... et cetera */
  /* Note: the below code is not very obvious at initial glance, but I
           wrote it such that we do not perform unnecessary memory
           allocation/deallocation as the R and MATLAB codes do implictly */
  for(i = 1; i < N; i++)  {
    Y1i_new = &Y1i[i];
    Y2i_new = &Y2i[i];
    for(j = 1; j < N-i+1; j++)  {
      Y2i_new[j-1] = Y2i_new[j-1] - Y2[i-1];
      Y1i_new[j-1] = Y1i_new[j-1] - Y1[i-1];
    }

    /* R code: G[i, i:n] <- Y2i - (Y1i*Y1i)/(1:(n-i+1)) */
    for(j = 0; j < N-i; j++)
      G[i + nrV[i+j]] = Y2i_new[j] - Y1i_new[j]*Y1i_new[j]/(j+1);
  }

  /* Free Y2, Y2i, and Y1i */
  Free(Y2);
  Free(Y2i);
  Free(Y1i);

  /* R code: I[1, ] <- G[1, ] (only after we fill in values for G matrix) */
  for(j = 0; j < N; j++)
    I[0 + (*Km)*j] = G[0 + nrV[j]];

  if(*Km > 2)  {
    for(k = 1; k < *Km-1; k++)  {
      for(L = k; L < N; L++)  {
	/* find the minimum */
	minVal = I[(k-1) + (*Km)*0] + G[(0+1) + nrV[L]];
	minIndex = 0;
	for(i = 1; i < L; i++)  {
	  temp = I[(k-1) + (*Km)*i] + G[(i+1) + nrV[L]];
	  if(temp < minVal)  {
	    minIndex = i;
	    minVal = temp;
	  }
	}

	I[k + (*Km)*L] = minVal;
	t[k-1 + (*Km-1)*L] = minIndex;
      }
    }
  }

  /* find the minimum */
  minVal = I[(*Km-2) + (*Km)*0] + G[(0+1) + nrV[N-1]];
  minIndex = 0;
  for(i = 1; i < N-1; i++)  {
    temp = I[(*Km-2) + (*Km)*i] + G[(i+1) + nrV[N-1]];
    if(temp < minVal)  {
      minIndex = i;
      minVal = temp;
    }
  }

  I[(*Km-1) + (*Km)*(N-1)] = minVal;
  t[(*Km-2) + (*Km-1)*(N-1)] = minIndex;

  /* R code: J <- I[ , n]     */
  /*         J <- log(J / n)  */
  for(i = 0; i < *Km; i++)
    J[i] = log(I[i + (*Km)*(N-1)] / N);

  for(i = 0; i < *Km; i++)
    th[i + (*Km)*i] = N-1;
  
  for(K = 1; K < *Km; K++) {
    for(k = K-1; k >= 0; k--)
      th[K + (*Km)*k] = t[k + (*Km-1)*(th[K + (*Km)*(k+1)])];
  }
  
  /* Convert contents of "th" to 1-based indices*/
  for(j = 0; j < *Km; j++)  {
    for(i = j; i < *Km; i++)
      th[i + (*Km)*j]++;
  }

  /* Free G, I, t matrices */
  Free(G);
  Free(nrV);
  Free(I);
  Free(t);

  return;
}
